(*
 *  emlStart.sml: This loads everything nessecary to get EML up and running.
 *)
val _ = print("Loading EML.\n");

(* 
 *  Turn off garbage collection messages.
 *)
SMLofNJ.Internals.GC.messages(false);

(*
 *  emlSettings_loadVerbose: Used for debugging.  When enabled (1), regular
 *      SML messages are printed as EML is loaded.  If disabled (0), these 
 *      messages are surpressed.
 *
 *  emlSettings_runVerbose: Used for debugging.  When enabled (1), additional
 *      EML messages are printed while running EML.  If disabled (0), only 
 *      basic information is printed.
 *)
val emlSettings_loadVerbose = 1;
val emlSettings_runVerbose = ref 1;

(*Compiler.Control.Print.out := {say=fn _=>(), flush=fn()=>()};*)

(* 
 *  If we have problems loading the MiniEML interpreter files, we raise
 *  a LoadError, with a string explaining the error.
 *)
exception LoadError of string;

(*
 *  emlPrint(): If emlSettings-runVerbose is set (see above), then the message
 *  will be printed.  Otherwise it will not be.
 *)
fun emlSetVerbose(x) = emlSettings_runVerbose := x

fun emlPrint(x) = 
    if !emlSettings_runVerbose = 1 then print(x)
    else ();

(* 
 *  Need to load in the yacc-related libararies and random utilities. 
 *)
let val cur_dir = OS.FileSys.getDir()
(*
    val cur_os = OS.Process.getEnv("OSTYPE")

    val os_name = case cur_os of
		       NONE => raise LoadError(
				"No OSTYPE environment variable defined.")
		     | SOME("linux") => "linux"
		     | SOME("linux-gnu") => "linux"
		     | SOME("solaris") => "solaris"
		     | SOME(unk_name) => "ERROR"
 *)
in (OS.FileSys.chDir("ml-yacc/lib");
    CM.make'("sources.cm");
    OS.FileSys.chDir(cur_dir))
end;

(* 
 *  Now actually include all the necessary files.
 *)

(* 
 *  Include the association list helper code.
 *)
use "emlalist.sml";

(*
 *  Load the datatypes and types first.
 *)
val _ = print("Loading internal structures.\n");
use "emlStructs.sml";
use "emlStructs_print.sml";

(* 
 *  Include the environment code.
 *)
val _ = print("Loading environment control.\n");
use "emlEnv.sml";

(* 
 *  Load the class and type specific information.
 *)
val _ = print("Loading class and type control.\n");
use "emlICEInit.sml";
use "emlClass.sml";
use "emlType.sml";

(* 
 *  Include the generated parser and lexer.
 *)
val _ = print("Loading parser and lexer.\n");
use "ml.mlyacc.sig";
use "ml.mllex.sml";
use "ml.mlyacc.sml";
	 
(* 
 *  Include the parser interface.
 *)
use "emlParser.sml";

(* 
 *  Include the evaluator code.
 *)
val _ = print("Loading ICE.\n");
use "emlInternalize.sml";
use "emlTypeCheck.sml";
use "emlEval.sml";
use "emlICEControl.sml";

(* 
 *  Now start things up.
 *)
val _ = print("Loading interactive interface.\n");
use "emlRepl.sml";
repl();
