(* 
 *  emlParser.sml: Gets everything working with the parser and lexer.
 *)

(* 
 *  The interface to the automatically generated parser and lexer.  You
 *  don't need to read this code. 
 *)

(* 
 *  Connect up the generated lexer and parser.
 *)
structure MLLrVals = 
    MLLrValsFun(structure Token = LrParser.Token)
structure MLLex =
    MLLexFun(structure Tokens = MLLrVals.Tokens)
structure MLParser =
    Join(structure ParserData = MLLrVals.ParserData
	 structure Lex = MLLex
	 structure LrParser = LrParser)

(* 
 *  The ParseError exception is raised if there is a parse error 
 *)
exception ParseError of string

(* 
 *  Parse the input string into a TopLevel AST, raising ParseError if
 *  there was a syntax error 
 *)
fun parse(inputString:string):TopLevel =
    let
(* 
 *      A hackish way of providing a pseudo-stream using the
 *      current string. 
 *)
        val nextLine:string ref = ref inputString
        fun stringStream(_) =
            let val currentLine = !nextLine
	    in (nextLine := ""; currentLine) end
        val lexer = MLParser.makeLexer(stringStream)
        fun print_error(msg:string,line:int,_) =
	    raise ParseError("line " ^ Int.toString(line) ^ ": " ^ msg)
        val (result,_) = MLParser.parse(0, lexer, print_error, ())
            handle MLLex.UserDeclarations.LexerError(msg:string) => 
	        raise ParseError(msg)
    in
        result
    end
