(*
 *  emlType.sml: Type specific code for EML.
 *
 *  This file could possibly be removed in the future, moving the code 
 *  elsewhere.
 *)

(*
 *  Takes a 
 *)

fun itcenv_addElem(nil, frontList, indexName, newPat) =
        raise TypeError("No message to add method pattern to. (ITC)")
  | itcenv_addElem((head as {name=msgName, gblType=headType, pats=patList})::rest,
        frontList, indexName, newPat) =
    if (msgName = indexName) then
        frontList @ {name=msgName, gblType=headType, pats=newPat::patList} :: rest
    else
        itcenv_addElem(rest, frontList @ [head], indexName, newPat)

(*
 *  Type datatype has been moved to emlStructures.sml.
 *)

fun isValidClassType(RecordType(_)) = true
  | isValidClassType(_) = false

fun isValidClassTypeOpt(SOME(t)) = isValidClassType(t)
  | isValidClassTypeOpt(NONE) = false

fun isValidFunType(FunctionType(_,_)) = true
  | isValidFunType(_) = false

fun coalesceClassTypes(RecordType(sRecList), RecordType(lRecList)) =
    let
(*
 *      Inserts one element into an alphabetized list.
 *)   
        fun alphabetizeInsert(elt, eltsBefore, nil) = eltsBefore @ [elt]
          | alphabetizeInsert(elt as {tident=eltID, ttype=eltType}, 
                eltsBefore, (eltCheck as {tident=checkID, ttype=checkType})::eltsRest) =
            if (eltID < checkID) then eltsBefore @ [elt] @ (eltCheck::eltsRest)
            else alphabetizeInsert(elt, eltsBefore @ [eltCheck], eltsRest)

(*
 *      Creates an alphabetized list from the input list
 *)             
        fun alphabetizeList(nil, newList) =  newList
          | alphabetizeList(firstRec::restRecs, newList) =
            let
                val updList = alphabetizeInsert(firstRec, nil, newList)
            in
                alphabetizeList(restRecs, updList)
            end
    in
(*
 *      Alphabetize the local class record list into the super class.
 *)    
         RecordType(alphabetizeList(lRecList, sRecList))
    end

  | coalesceClassTypes(RecordType(_), second) = raise TypeError("Class has " ^
                        "non-record type: " ^ printType(second) ^ ".")
  | coalesceClassTypes(first, RecordType(_)) = raise TypeError("Superclass has " ^
                        "non-record type: " ^ printType(first) ^ ".")
  | coalesceClassTypes(first, second) = raise TypeError("Class has non-record " ^
                        "type: " ^ printType(second) ^ " and superclass has " ^
                        "non-record type: " ^ printType(first) ^ ".")

(*
 *  Used in emlTypeCheck.sml to get the pattern and body from a function type.
 *)
fun typeGetFunParts(FunctionType(pat, body)) = (pat, body)
  | typeGetFunParts(t) = raise TypeError("Type (" ^ printType(t) ^ " is not "
                             ^ "a valid class type.")

fun typeGetBinopArgs(TupleType([arg1_type, arg2_type])) = (arg1_type, arg2_type)
  | typeGetBinopArgs(TupleType(_)) = raise TypeError("Binop does not have 2 " ^
                                               "arguments (although it should)")
  | typeGetBinopArgs(_) = raise TypeError("Binop is not a tuple type!")

fun typeGetTypesFromTuple(TupleType(tt_list)) = tt_list
  | typeGetTypesFromTuple(t) = raise TypeError("Type: " ^ printType(t) ^ 
        " isn't a tuple type.")

fun typeGetTypeFromList(ListType(list_type)) = list_type
  | typeGetTypeFromList(t) = raise TypeError("Type: " ^ printType(t) ^
        " isn't a list type.")

fun typeCreateListType(h_type, t_type) =
    case (t_type) of
        ListType(t_list_type) => if (t_list_type = h_type) then ListType(h_type)
                                 else raise TypeError("Tail elements don't " ^
                                                "have the same type as head.")
      | _ => raise TypeError("Tail must have list type.")

(* 
 *  Used in emlEval.sml to compare types.
 *)
fun compareSpecificity(pattern1, pattern2) = 
(* 
 *  If the two patterns are both as specific as each other, then they have the
 *  same specificity...
 *)
    if FirstIsAsSpecificAsSecond(pattern1, pattern2) andalso
       FirstIsAsSpecificAsSecond(pattern2, pattern1) then
        SameSpecificity
    else
(*
 *  If the first as specific as the second, and the second is not as specific as
 *  the first, then the first is more specific.
 *)
        if FirstIsAsSpecificAsSecond(pattern1, pattern2) andalso
           not(FirstIsAsSpecificAsSecond(pattern2, pattern1)) then
            FirstIsMoreSpecific
        else
            if FirstIsAsSpecificAsSecond(pattern2, pattern1) andalso
               not(FirstIsAsSpecificAsSecond(pattern1, pattern2)) then
                SecondIsMoreSpecific
            else
                Incomperable

(*
 *  Returns true if the first is at least as specific as the second, otherwise
 *  returns false
 *)
and FirstIsAsSpecificAsSecond(_,WildcardPatternIntern) = true
  | FirstIsAsSpecificAsSecond(_,VarPatternIntern(_)) = true
  | FirstIsAsSpecificAsSecond(IntPatternIntern(x),IntPatternIntern(y)) = x=y
  | FirstIsAsSpecificAsSecond(RealPatternIntern(x), RealPatternIntern(y)) = 
        x <= y andalso x >= y
  | FirstIsAsSpecificAsSecond(StringPatternIntern(x),StringPatternIntern(y)) = x=y
  | FirstIsAsSpecificAsSecond(BoolPatternIntern(x),BoolPatternIntern(y)) = x=y
  | FirstIsAsSpecificAsSecond(NilPatternIntern,NilPatternIntern) = true
  | FirstIsAsSpecificAsSecond(TuplePatternIntern(l1),TuplePatternIntern(l2)) = 
        AllFirstsAreAsSpecificAsAllSeconds(l1,l2)
  | FirstIsAsSpecificAsSecond(RecordPatternIntern(r1), RecordPatternIntern(r2)) =
    let
        fun recCheckOneElt({ident=id1, pattern=pat1}, {ident=id2, pattern=pat2}) =
            if (id1 = id2) then FirstIsAsSpecificAsSecond(pat1, pat2)
            else false

        fun recCheckAllElts(nil, nil) = true
          | recCheckAllElts(nil, notNil) = false
          | recCheckAllElts(notNil, nil) = false
          | recCheckAllElts(hd1::tl1, hd2::tl2) =
                recCheckOneElt(hd1, hd2) andalso recCheckAllElts(tl1, tl2)
    in
        recCheckAllElts(r1, r2)
    end

  | FirstIsAsSpecificAsSecond(ConsPatternIntern({head=h1,tail=t1}),
                              ConsPatternIntern({head=h2,tail=t2})) = 
        FirstIsAsSpecificAsSecond(h1,h2) andalso 
        FirstIsAsSpecificAsSecond(t1,t2)
  | FirstIsAsSpecificAsSecond(ClassPatternIntern({class=c1,rep=r1}),
                              ClassPatternIntern({class=c2,rep=r2})) =
        if FirstIsStrictSubclassOfSecond(c1,c2) then true
        else 
            if areSameClass(c1,c2) andalso FirstIsAsSpecificAsSecond(r1,r2) 
            then true
            else false
  | FirstIsAsSpecificAsSecond(_,_) = false

and AllFirstsAreAsSpecificAsAllSeconds(x1::x1s,x2::x2s) = 
        FirstIsAsSpecificAsSecond(x1,x2) andalso 
        AllFirstsAreAsSpecificAsAllSeconds(x1s,x2s)
  | AllFirstsAreAsSpecificAsAllSeconds(nil,nil) = true
  | AllFirstsAreAsSpecificAsAllSeconds(nil,_) = false
  | AllFirstsAreAsSpecificAsAllSeconds(_,nil) = false

(*
 * Given a type, returns the number of class types that have been
 * declared as owner classes.
 *)
and OwnerTypeCount(IntegerType) = 0
  | OwnerTypeCount(RealType) = 0
  | OwnerTypeCount(BooleanType) = 0
  | OwnerTypeCount(StringType) = 0
  | OwnerTypeCount(TupleType(type_list)) = OwnerTypeListCount(type_list)
  | OwnerTypeCount(ListType(list_type)) = OwnerTypeCount(list_type)
  | OwnerTypeCount(UnevaluatedClassType{class_name, is_owner=NonOwnerClass}) = 0
  | OwnerTypeCount(UnevaluatedClassType{class_name, is_owner=OwnerClass}) = 1
  | OwnerTypeCount(ClassInstanceType{class, is_owner=NonOwnerClass}) = 0
  | OwnerTypeCount(ClassInstanceType{class, is_owner=OwnerClass}) = 1
  | OwnerTypeCount(RecordType(rec_list)) = OwnerTypeRecordCount(rec_list)
  | OwnerTypeCount(FunctionType(arg_type, return_type)) =
        OwnerTypeCount(arg_type) + OwnerTypeCount(return_type)

and OwnerTypeListCount(nil) = 0
  | OwnerTypeListCount(head_type :: tail_types) =
        OwnerTypeCount(head_type) + OwnerTypeListCount(tail_types)

and OwnerTypeRecordCount(nil) = 0
  | OwnerTypeRecordCount({tident, ttype} :: tail_records) =
        OwnerTypeCount(ttype) + OwnerTypeRecordCount(tail_records)

(*
 *  Given a type, returns the class that is marked as the owner type.
 *  Raises NotFound if there is no ower type.
 *)
and findOwnerClassName(IntegerType) = raise NotFound
  | findOwnerClassName(RealType) = raise NotFound
  | findOwnerClassName(BooleanType) = raise NotFound
  | findOwnerClassName(StringType) = raise NotFound
  | findOwnerClassName(TupleType(type_list)) = findOwnerClassNameList(type_list)
  | findOwnerClassName(ListType(list_type)) = findOwnerClassName(list_type)
  | findOwnerClassName(UnevaluatedClassType{class_name, is_owner=NonOwnerClass}) =
        raise NotFound
  | findOwnerClassName(UnevaluatedClassType{class_name, is_owner=OwnerClass}) =
        class_name
  | findOwnerClassName(ClassInstanceType{class, is_owner=NonOwnerClass}) =
        raise NotFound
  | findOwnerClassName(ClassInstanceType{class=Class{name, id, mode, superclass,
        interfaces, dType}, is_owner=OwnerClass}) = name
  | findOwnerClassName(RecordType(rec_list)) = findOwnerClassNameRecord(rec_list)
  | findOwnerClassName(FunctionType(arg_type, return_type)) =
        findOwnerClassName(arg_type)
            handle NotFound => findOwnerClassName(return_type)

and findOwnerClassNameList(nil) = raise NotFound
  | findOwnerClassNameList(head_type :: tail_types) =
        findOwnerClassName(head_type)
            handle NotFound => findOwnerClassNameList(tail_types)

and findOwnerClassNameRecord(nil) = raise NotFound
  | findOwnerClassNameRecord({tident, ttype} :: tail_records) =
        findOwnerClassName(ttype)
            handle NotFound => findOwnerClassNameRecord(tail_records)
